# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator (autorest: 3.9.7, generator: @autorest/python@6.12.3)
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from ._operations import (
    AuditproceduresOperations,
    AuditsOperations,
    DatasetsOperations,
    EvaluationsOperations,
    InferenceservicesOperations,
    InferencesessionsOperations,
    ModelsOperations,
    ReportsOperations,
)
from ._patch import *  # pylint: disable=unused-wildcard-import
from ._patch import __all__ as _patch_all
from ._patch import patch_sdk as _patch_sdk

__all__ = [
    "AuditsOperations",
    "AuditproceduresOperations",
    "DatasetsOperations",
    "EvaluationsOperations",
    "InferenceservicesOperations",
    "InferencesessionsOperations",
    "ModelsOperations",
    "ReportsOperations",
]
__all__.extend([p for p in _patch_all if p not in __all__])
_patch_sdk()
